@@ -1,20 +1,57 @@
 package JSON::Validator::Schema::OpenAPIv3;
 use Mojo::Base 'JSON::Validator::Schema::Draft201909';
 
-use JSON::Validator::Schema::OpenAPIv2;
 use JSON::Validator::Util qw(E data_type negotiate_content_type schema_type);
-use Mojo::JSON qw(false true);
+use Mojo::JSON            qw(false true);
 use Mojo::Path;
-use Mojo::Util qw(monkey_patch);
 
 has moniker       => 'openapiv3';
 has specification => 'https://spec.openapis.org/oas/3.0/schema/2019-04-02';
 
-# some methods are shared with OpenAPIv2
-monkey_patch __PACKAGE__,
-  $_ => JSON::Validator::Schema::OpenAPIv2->can($_)
-  for qw(coerce routes validate_request validate_response),
-  qw(_coerce_arrays _coerce_default_value _find_all_nodes _prefix_error_path _validate_request_or_response);
+require JSON::Validator::Schema::OpenAPIv2;
+*coerce                           = \&JSON::Validator::Schema::OpenAPIv2::coerce;
+*errors                           = \&JSON::Validator::Schema::OpenAPIv2::errors;
+*routes                           = \&JSON::Validator::Schema::OpenAPIv2::routes;
+*validate_request                 = \&JSON::Validator::Schema::OpenAPIv2::validate_request;
+*validate_response                = \&JSON::Validator::Schema::OpenAPIv2::validate_response;
+*_coerce_arrays                   = \&JSON::Validator::Schema::OpenAPIv2::_coerce_arrays;
+*_coerce_default_value            = \&JSON::Validator::Schema::OpenAPIv2::_coerce_default_value;
+*_find_all_nodes                  = \&JSON::Validator::Schema::OpenAPIv2::_find_all_nodes;
+*_params_for_add_default_response = \&JSON::Validator::Schema::OpenAPIv2::_params_for_add_default_response;
+*_prefix_error_path               = \&JSON::Validator::Schema::OpenAPIv2::_prefix_error_path;
+*_validate_request_or_response    = \&JSON::Validator::Schema::OpenAPIv2::_validate_request_or_response;
+
+sub add_default_response {
+  my ($self, $params) = ($_[0], shift->_params_for_add_default_response(@_));
+
+  my $schemas = $self->data->{components}{schemas} ||= {};
+  $schemas->{$params->{name}} ||= $params->{schema};
+  my $ref = {'$ref' => "#/components/schemas/$params->{name}"};
+  $self->_register_ref($ref, schema => $schemas->{$params->{name}});
+
+  for my $route ($self->routes->each) {
+    my $op = $self->get([paths => @$route{qw(path method)}]);
+    for my $status (@{$params->{status}}) {
+      next if $self->get(['paths', @$route{qw(path method)}, 'responses', $status]);
+      $op->{responses}{$status}{content}{'application/json'} //= {schema => $ref};
+      $op->{responses}{$status}{description} //= $params->{description};
+    }
+  }
+
+  return $self;
+}
+
+sub base_url {
+  my ($self, $url) = @_;
+
+  # Get
+  return Mojo::URL->new($self->get('/servers/0/url') || '') unless $url;
+
+  # Set
+  $url = Mojo::URL->new($url)->to_abs($self->base_url);
+  $self->data->{servers}[0]{url} = $url->to_string;
+  return $self;
+}
 
 sub new {
   my $self = shift->SUPER::new(@_);
@@ -31,10 +68,6 @@ sub parameters_for_request {
   return undef unless $self->get([paths => $path, $method]);
 
   my @parameters = map {@$_} $self->_find_all_nodes([paths => $path, $method], 'parameters');
-  for my $param (@parameters) {
-    $param->{type} ||= schema_type($param->{schema});
-  }
-
   if (my $request_body = $self->get([paths => $path, $method, 'requestBody'])) {
     my @accepts = sort keys %{$request_body->{content} || {}};
     push @parameters,
@@ -58,8 +91,8 @@ sub parameters_for_response {
   my $cache_key = "parameters_for_response:$method:$path:$status";
   return $self->{cache}{$cache_key} if $self->{cache}{$cache_key};
 
-  my $responses = $self->get([paths => $path, $method, 'responses']);
-  my $response  = $responses->{$status} || $responses->{default};
+  my $response = $self->get([paths => $path, $method, 'responses', $status])
+    || $self->get([paths => $path, $method, 'responses', 'default']);
   return undef unless $response;
 
   my @parameters;
@@ -107,15 +140,20 @@ sub _build_formats {
   };
 }
 
+sub _bundle_ref_path_expand {
+  my ($self, $schema, $ref) = @_;
+  return $ref =~ m!\bcomponents/([^/]+)/(.+)! ? ('components', $1, $2) : ('components', 'schemas', $ref);
+}
+
 sub _coerce_parameter_format {
   my ($self, $val, $param) = @_;
   return unless $val->{exists};
 
   state $in_style = {cookie => 'form', header => 'simple', path => 'simple', query => 'form'};
   $param->{style} = $in_style->{$param->{in}} unless $param->{style};
-  return $self->_coerce_parameter_style_object_deep($val, $param) if $param->{style} eq 'deepObject';
+  return $self->_coerce_parameter_style_object_deep($val, $param) if +($param->{style} // '') eq 'deepObject';
 
-  my $schema_type = schema_type $param;
+  my $schema_type = schema_type $param->{schema};
   return $self->_coerce_parameter_style_array($val, $param)  if $schema_type eq 'array';
   return $self->_coerce_parameter_style_object($val, $param) if $schema_type eq 'object';
 }
@@ -148,7 +186,7 @@ sub _coerce_parameter_style_object {
   my $explode = $param->{explode} // (grep { $style eq $_ } qw(cookie query)) ? 1 : 0;
 
   if ($explode) {
-    return if $style eq 'form';
+    return $self->_coerce_parameter_style_object_form($val, $param) if $style eq 'form';
     state $style_re = {label => qr{\.}, matrix => qr{;}, simple => qr{,}};
     return unless my $re = $style_re->{$style};
     return if $style eq 'matrix' && $val->{value} !~ s/^;//;
@@ -204,15 +242,19 @@ sub _coerce_parameter_style_object_deep {
   return $val->{exists} = 0;
 }
 
-sub _definitions_path_for_ref {
-  my ($self, $ref) = @_;
-  my $path = Mojo::Path->new($ref->fqn =~ m!^.*#/(components/.+)$!)->to_dir->parts;
-  return $path->[0] ? $path : ['definitions'];
+sub _coerce_parameter_style_object_form {
+  my ($self, $val, $param) = @_;
+  return unless my $properties = $param->{schema} && $param->{schema}{properties};
+
+  for my $k (keys %{$val->{value}}) {
+    next unless my $type = $properties->{$k} && $properties->{$k}{type};
+    $val->{value}{$k} = [$val->{value}{$k}] if $type eq 'array' and ref $val->{value}{$k} ne 'ARRAY';
+  }
 }
 
 sub _get_parameter_value {
   my ($self, $param, $get) = @_;
-  my $schema_type = schema_type $param;
+  my $schema_type = schema_type $param->{schema};
   my $name        = $param->{name};
   $name = undef if $schema_type eq 'object' && $param->{explode} && ($param->{style} || '') =~ m!^(form|deepObject)$!;
 
@@ -231,21 +273,25 @@ sub _to_list { ref $_[0] eq 'ARRAY' ? @{$_[0]} : $_[0] ? ($_[0]) : () }
 
 sub _validate_body {
   my ($self, $direction, $val, $param) = @_;
-  $val->{content_type} = $param->{accepts}[0] if !$val->{content_type} and @{$param->{accepts}};
 
   if ($val->{accept}) {
     $val->{content_type} = negotiate_content_type($param->{accepts}, $val->{accept});
     $val->{valid}        = $val->{content_type} ? 1 : 0;
     return E "/header/Accept", [join(', ', @{$param->{accepts}}), type => $val->{accept}] unless $val->{valid};
   }
-  if (@{$param->{accepts}} and !$val->{content_type}) {
-    $val->{valid} = 0;
-    return E "/$param->{name}", [join(', ', @{$param->{accepts}}) => type => $val->{content_type}];
+  if (@{$param->{accepts}} and $val->{content_type}) {
+    my $negotiated = negotiate_content_type($param->{accepts}, $val->{content_type});
+    $val->{valid} = $negotiated ? 1 : 0;
+    return E "/$param->{name}", [join(', ', @{$param->{accepts}}) => type => $val->{content_type}] unless $negotiated;
   }
   if ($param->{required} and !$val->{exists}) {
+    $val->{valid} = 0;
     return E "/$param->{name}", [qw(object required)];
   }
   if ($val->{exists}) {
+    $val->{content_type} //= $param->{accepts}[0];
+    local $self->{coerce}{arrays} = 1
+      if $val->{content_type} =~ m!^(application/x-www-form-urlencoded|multipart/form-data)$!;
     local $self->{"validate_$direction"} = 1;
     my @errors = map { $_->path(_prefix_error_path($param->{name}, $_->path)); $_ }
       $self->validate($val->{value}, $param->{content}{$val->{content_type}}{schema});
@@ -256,8 +302,38 @@ sub _validate_body {
   return;
 }
 
+sub _validate_id { }
+
+sub _validate_type_array {
+  my $self = shift;
+  $_[0] = [$_[0]] if ref $_[0] ne 'ARRAY' and $self->{coerce}{arrays};
+  return $_[1]->{schema}{nullable} && !defined $_[0] ? () : $self->SUPER::_validate_type_array(@_);
+}
+
+sub _validate_type_boolean {
+  my $self = shift;
+  return $_[1]->{schema}{nullable} && !defined $_[0] ? () : $self->SUPER::_validate_type_boolean(@_);
+}
+
+sub _validate_type_enum {
+  my $self = shift;
+  return $_[1]->{schema}{nullable} && !defined $_[0] ? () : $self->SUPER::_validate_type_enum(@_);
+}
+
+sub _validate_type_integer {
+  my $self = shift;
+  return $_[1]->{schema}{nullable} && !defined $_[0] ? () : $self->SUPER::_validate_type_integer(@_);
+}
+
+sub _validate_type_number {
+  my $self = shift;
+  return $_[1]->{schema}{nullable} && !defined $_[0] ? () : $self->SUPER::_validate_type_number(@_);
+}
+
 sub _validate_type_object {
-  my ($self, $data, $path, $schema) = @_;
+  my ($self, $data, $state) = @_;
+  my ($path, $schema) = @$state{qw(path schema)};
+  return if $schema->{nullable} && !defined $data;
   return E $path, [object => type => data_type $data] if ref $data ne 'HASH';
   return shift->SUPER::_validate_type_object(@_) unless $self->{validate_request} or $self->{validate_response};
 
@@ -268,31 +344,23 @@ sub _validate_type_object {
     return E $path, "Discriminator $name has no value."          unless my $map_name = $data->{$name};
     return E $path, "No definition for discriminator $map_name." unless my $url      = $mapping->{$map_name};
     return E $path, "TODO: Not yet supported: $url"              unless $url =~ s!^#!!;
-    local $self->{inside_discriminator} = 1;    # prevent recursion
-    return $self->_validate($data, $path, $self->get($url));
-  }
-
-  my %properties = %{$schema->{properties} || {}};
-  local $schema->{properties} = \%properties;
-  for my $key (keys %properties) {
-    next unless $properties{$key}{nullable};
-    my $tied = tied %{$properties{$key}};
-    $properties{$key} = $tied ? {%{$tied->schema}} : {%{$properties{$key}}};
-    $properties{$key}{type} = ['null', _to_list($properties{$key}{type})];
+    local $self->{inside_discriminator} = 1;
+    return $self->_validate($data, $self->_state($state, schema => $self->get($url)));
   }
 
   return $self->{validate_request}
-    ? $self->_validate_type_object_request($_[1], $path, $schema)
-    : $self->_validate_type_object_response($_[1], $path, $schema);
+    ? $self->_validate_type_object_request($_[1], $state)
+    : $self->_validate_type_object_response($_[1], $state);
 }
 
 sub _validate_type_object_request {
-  my ($self, $data, $path, $schema) = @_;
+  my ($self, $data, $state) = @_;
+  my ($path, $schema) = @$state{qw(path schema)};
 
   my (@errors, %ro);
   for my $name (keys %{$schema->{properties} || {}}) {
-    next unless $schema->{properties}{$name}{readOnly};
-    push @errors, E "$path/$name", "Read-only." if exists $data->{$name};
+    next unless $self->_get(['readOnly'], {path => [], schema => $schema->{properties}{$name}});
+    push @errors, E [@$path, $name], "Read-only." if exists $data->{$name};
     $ro{$name} = 1;
   }
 
@@ -300,19 +368,20 @@ sub _validate_type_object_request {
 
   return (
     @errors,
-    $self->_validate_type_object_min_max($_[1], $path, $schema),
-    $self->_validate_type_object_dependencies($_[1], $path, $schema),
-    $self->_validate_type_object_properties($_[1], $path, $schema),
+    $self->_validate_type_object_min_max($_[1], $state),
+    $self->_validate_type_object_dependencies($_[1], $state),
+    $self->_validate_type_object_properties($_[1], $state),
   );
 }
 
 sub _validate_type_object_response {
-  my ($self, $data, $path, $schema) = @_;
+  my ($self, $data, $state) = @_;
+  my ($path, $schema) = @$state{qw(path schema)};
 
   my (@errors, %rw);
   for my $name (keys %{$schema->{properties} || {}}) {
-    next unless $schema->{properties}{$name}{writeOnly};
-    push @errors, E "$path/$name", "Write-only." if exists $data->{$name};
+    next unless $self->_get(['writeOnly'], {path => [], schema => $schema->{properties}{$name}});
+    push @errors, E [@$path, $name], "Write-only." if exists $data->{$name};
     $rw{$name} = 1;
   }
 
@@ -320,12 +389,17 @@ sub _validate_type_object_response {
 
   return (
     @errors,
-    $self->_validate_type_object_min_max($_[1], $path, $schema),
-    $self->_validate_type_object_dependencies($_[1], $path, $schema),
-    $self->_validate_type_object_properties($_[1], $path, $schema),
+    $self->_validate_type_object_min_max($_[1], $state),
+    $self->_validate_type_object_dependencies($_[1], $state),
+    $self->_validate_type_object_properties($_[1], $state),
   );
 }
 
+sub _validate_type_string {
+  my $self = shift;
+  return $_[1]->{schema}{nullable} && !defined $_[0] ? () : $self->SUPER::_validate_type_string(@_);
+}
+
 1;
 
 =encoding utf8
@@ -344,6 +418,12 @@ This class represents L<https://spec.openapis.org/oas/3.0/schema/2019-04-02>.
 
 =head1 ATTRIBUTES
 
+=head2 errors
+
+  my $array_ref = $schema->errors;
+
+See L<JSON::Validator::Schema/errors>.
+
 =head2 moniker
 
   $str    = $schema->moniker;
@@ -360,6 +440,22 @@ Defaults to "L<https://spec.openapis.org/oas/3.0/schema/2019-04-02>".
 
 =head1 METHODS
 
+=head2 add_default_response
+
+  $schema = $schema->add_default_response(\%params);
+
+See L<JSON::Validator::Schema::OpenAPIv2/add_default_response> for details.
+
+=head2 base_url
+
+  $url = $schema->base_url;
+  $schema = $schema->base_url($url);
+
+Can get or set the default URL for this schema. C<$url> can be either a
+L<Mojo::URL> object or a plain string.
+
+This method will read or write "/servers/0/url" in L</data>.
+
 =head2 coerce
 
   my $schema   = $schema->coerce({booleans => 1, numbers => 1, strings => 1});
